print("import: 'debugpy'")
import debugpy

print("import: 'debugpy.adapter'")
import debugpy.adapter

print("import: 'debugpy.common'")
import debugpy.common

print("import: 'debugpy.launcher'")
import debugpy.launcher

print("import: 'debugpy.server'")
import debugpy.server

